@@ -2,7 +2,7 @@
 # -*- Mode: Python; coding: utf-8; indent-tabs-mode: nil; tab-width: 4 -*-
 #   Catfish - a versatile file searching tool
 #   Copyright (C) 2007-2012 Christian Dywan <christian@twotoasts.de>
-#   Copyright (C) 2012-2016 Sean Davis <smd.seandavis@gmail.com>
+#   Copyright (C) 2012-2019 Sean Davis <smd.seandavis@gmail.com>
 #
 #   This program is free software: you can redistribute it and/or modify it
 #   under the terms of the GNU General Public License version 2, as published
@@ -14,12 +14,13 @@
 #   PURPOSE.  See the GNU General Public License for more details.
 #
 #   You should have received a copy of the GNU General Public License along
-#   with this program.  If not, see <http://www.gnu.org/licenses/>.
+#   with this program.  If not, see <https://www.gnu.org/licenses/>.
 
 import logging
 
 import io
 import os
+import re
 import signal
 import subprocess
 import time
@@ -50,11 +51,21 @@ else:
     locate_support = False
 FNULL.close()
 
+def get_keyword_list(keywords):
+    keywords = keywords.replace(",", " ").strip().lower()
+    kwords = []
+    matches = re.findall(r'\"(.+?)\"', keywords)
+    for match in matches:
+        newmatch = match.replace(" ", "\0")
+        newmatch = newmatch.replace("\"", "")
+        keywords = keywords.replace("\"%s\"" % match, newmatch)
+    for keyword in keywords.split(" "):
+        kwords.append(keyword.replace("\0", " "))
+    return kwords
 
 def string_regex(keywords, path):
     """Returns a string with the regular expression containing all combinations
     of the keywords."""
-    keywords = keywords.split()
     if len(keywords) == 0:
         return ''
     if len(keywords) == 1:
@@ -81,7 +92,7 @@ class CatfishSearchEngine:
     """CatfishSearchEngine is the collection of search backends that are used
     to perform a query.  Each backend is a CatfishSearchMethod"""
 
-    def __init__(self, methods=['zeitgeist', 'walk']):
+    def __init__(self, methods=['zeitgeist', 'locate', 'walk']):
         """Initialize the CatfishSearchEngine.  Provide a list of methods to
         be included in the search backends.  Available backends include:
 
@@ -122,7 +133,7 @@ class CatfishSearchEngine:
         """Add a CatfishSearchMethod the the engine's search backends."""
         self.methods.append(method_class())
 
-    def run(self, keywords, path, limit=-1, regex=False):
+    def run(self, keywords, path, limit=-1, regex=False): # noqa
         """Run the CatfishSearchEngine.
 
         Each method is run sequentially in the order they are added.
@@ -131,37 +142,59 @@ class CatfishSearchEngine:
         filename is yielded if it matches the query.  False is also yielded
         afterwards to guarantee the interface does not lock up."""
         self.start_time = time.time()
-        keywords = keywords.replace(',', ' ').strip().lower()
+        self.stop_time = 0
+
+        keywords = get_keyword_list(keywords)
+        self.keywords = " ".join(keywords)
 
         logger.debug("[%i] path: %s, keywords: %s, limit: %i, regex: %s",
                      self.engine_id, str(path), str(keywords), limit,
                      str(regex))
 
-        self.keywords = keywords
-
         wildcard_chunks = []
-        for key in self.keywords.split():
+        keys = []
+        for key in keywords:
             if '*' in key:
                 wildcard_chunks.append(key.split('*'))
-
-        keywords = keywords.replace('*', ' ')
+            else:
+                keys.append(key)
 
         # For simplicity, make sure the path contains a trailing '/'
         if not path.endswith('/'):
             path += '/'
 
-        # Transform the keywords into a clean list.
-        keys = []
-        for key in keywords.split():
-            keys.append(key.strip())
+        # Path exclusions for efficiency
+        exclude = []
+        maybe_exclude = [
+            os.path.expanduser("~/.cache"),
+            os.path.expanduser("~/.gvfs"),
+            "/dev"
+        ]
+        for maybe_path in maybe_exclude:
+            if not path.startswith(maybe_path):
+                exclude.append(maybe_path)
 
         file_count = 0
         for method in self.methods:
+            if self.stop_time > 0:
+                logger.debug("Engine is stopped")
+                return
             logger.debug(
                 "[%i] Starting search method: %s",
                 self.engine_id, method.method_name)
             for filename in method.run(keywords, path, regex):
                 if isinstance(filename, str) and path in filename:
+                    found_bad = False
+                    for filepath in exclude:
+                        if filepath in filename:
+                            if self.stop_time > 0:
+                                logger.debug("Engine is stopped")
+                                return
+                            found_bad = True
+                    if found_bad:
+                        yield True
+                        continue
+
                     if method.method_name == 'fulltext' or  \
                             all(key in
                                 os.path.basename(filename).lower()
@@ -264,10 +297,21 @@ class CatfishSearchMethod_Walk(CatfishSearchMethod):
 
         This function is a generator and will yield files as they are found or
         True if still running."""
+        exclude = []
+        maybe_exclude = [
+            os.path.expanduser("~/.cache"),
+            os.path.expanduser("~/.gvfs"),
+            "/dev"
+        ]
+        for maybe_path in maybe_exclude:
+            if not path.startswith(maybe_path):
+                exclude.append(maybe_path)
+
         self.running = True
         if isinstance(keywords, str):
             keywords = keywords.replace(',', ' ').strip().split()
         for root, dirs, files in os.walk(path, False):
+            dirs[:] = [d for d in dirs if os.path.join(root, d) not in exclude]
             if not self.running:
                 break
             paths = dirs + files
@@ -300,7 +344,7 @@ class CatfishSearchMethod_Fulltext(CatfishSearchMethod):
         self.running = False
         self.exact = False
 
-    def run(self, keywords, path, regex=False):
+    def run(self, keywords, path, regex=False): # noqa
         """Run the search method using keywords and path.  regex is not used
         by this search method.
 
@@ -342,7 +386,7 @@ class CatfishSearchMethod_Fulltext(CatfishSearchMethod):
                                     break
 
                                 if self.exact:
-                                    if keywords in line:
+                                    if " ".join(keywords) in line:
                                         yield os.path.join(root, filename)
                                         break
                                 else:
@@ -468,7 +512,7 @@ class CatfishSearchMethodExternal(CatfishSearchMethod):
     def process_output(self, output):
         """Return the output text."""
         if isinstance(output, io.BufferedReader):
-            return map(lambda s: s.decode(encoding='UTF8').strip(),
+            return map(lambda s: s.decode(encoding='UTF8', errors='replace').strip(),
                        output.readlines())
         else:
             return output
@@ -503,8 +547,8 @@ class CatfishSearchMethod_Locate(CatfishSearchMethodExternal):
     def __init__(self):
         """Initialize the Locate SearchMethod."""
         CatfishSearchMethodExternal.__init__(self, "locate")
-        self.command = ["locate", "-i", "%keywords"]
+        self.command = ["locate", "-i", "%path*%keywords*"]
 
     def assemble_query(self, keywords, path):
         """Assemble the search query."""
-        return ["locate", "-i", str(keywords)]
+        return ["locate", "-i", "%path*", str(keywords)]
